package dev.fyloz.colorrecipesexplorer.model

import com.fasterxml.jackson.annotation.JsonIgnore
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.CannotDeleteException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import javax.persistence.*
import javax.validation.constraints.Min
import javax.validation.constraints.NotBlank


@Entity
@Table(name = "mix")
data class Mix(
    @Id
    @GeneratedValue(strategy = GenerationType.IDENTITY)
    override val id: Long?,

    var location: String?,

    @JsonIgnore
    @ManyToOne
    @JoinColumn(name = "recipe_id")
    val recipe: Recipe,

    @ManyToOne
    @JoinColumn(name = "mix_type_id")
    var mixType: MixType,

    @OneToMany(cascade = [CascadeType.ALL], fetch = FetchType.EAGER, orphanRemoval = true)
    @JoinColumn(name = "mix_id")
    var mixMaterials: MutableSet<MixMaterial>,
) : Model

open class MixSaveDto(
    @field:NotBlank
    val name: String,

    val recipeId: Long,

    val materialTypeId: Long,

    val mixMaterials: Set<MixMaterialDto>?
) : EntityDto<Mix>

open class MixUpdateDto(
    val id: Long,

    @field:NotBlank
    val name: String?,

    val materialTypeId: Long?,

    var mixMaterials: Set<MixMaterialDto>?
) : EntityDto<Mix>

data class MixOutputDto(
    val id: Long,
    val location: String?,
    val mixType: MixType,
    val mixMaterials: Set<MixMaterialOutputDto>
)

data class MixDeductDto(
    val id: Long,

    @field:Min(0, message = VALIDATION_SIZE_GE_ZERO)
    val ratio: Float
)

data class MixLocationDto(
    val mixId: Long,

    val location: String?
)

//fun Mix.toOutput() =

// ==== DSL ====
fun mix(
    id: Long? = null,
    location: String? = "location",
    recipe: Recipe = recipe(),
    mixType: MixType = mixType(),
    mixMaterials: MutableSet<MixMaterial> = mutableSetOf(),
    op: Mix.() -> Unit = {}
) = Mix(id, location, recipe, mixType, mixMaterials).apply(op)

fun mixSaveDto(
    name: String = "name",
    recipeId: Long = 0L,
    materialTypeId: Long = 0L,
    mixMaterials: Set<MixMaterialDto>? = setOf(),
    op: MixSaveDto.() -> Unit = {}
) = MixSaveDto(name, recipeId, materialTypeId, mixMaterials).apply(op)

fun mixUpdateDto(
    id: Long = 0L,
    name: String? = "name",
    materialTypeId: Long? = 0L,
    mixMaterials: Set<MixMaterialDto>? = setOf(),
    op: MixUpdateDto.() -> Unit = {}
) = MixUpdateDto(id, name, materialTypeId, mixMaterials).apply(op)

fun mixRatio(
    id: Long = 0L,
    ratio: Float = 1f,
    op: MixDeductDto.() -> Unit = {}
) = MixDeductDto(id, ratio).apply(op)

fun mixLocationDto(
    mixId: Long = 0L,
    location: String? = "location",
    op: MixLocationDto.() -> Unit = {}
) = MixLocationDto(mixId, location).apply(op)

// ==== Exceptions ====
private const val MIX_NOT_FOUND_EXCEPTION_TITLE = "Mix not found"
private const val MIX_ALREADY_EXISTS_EXCEPTION_TITLE = "Mix already exists"
private const val MIX_CANNOT_DELETE_EXCEPTION_TITLE = "Cannot delete mix"
private const val MIX_EXCEPTION_ERROR_CODE = "mix"

fun mixIdNotFoundException(id: Long) =
    NotFoundException(
        MIX_EXCEPTION_ERROR_CODE,
        MIX_NOT_FOUND_EXCEPTION_TITLE,
        "A mix with the id $id could not be found",
        id
    )

fun mixIdAlreadyExistsException(id: Long) =
    AlreadyExistsException(
        MIX_EXCEPTION_ERROR_CODE,
        MIX_ALREADY_EXISTS_EXCEPTION_TITLE,
        "A mix with the id $id already exists",
        id
    )

fun cannotDeleteMixException(mix: Mix) =
    CannotDeleteException(
        MIX_EXCEPTION_ERROR_CODE,
        MIX_CANNOT_DELETE_EXCEPTION_TITLE,
        "Cannot delete the mix ${mix.mixType.name} because one or more mixes depends on it"
    )
