package dev.fyloz.colorrecipesexplorer.service

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.repository.MixRepository
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.TestInstance
import kotlin.test.assertEquals
import kotlin.test.assertTrue

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
class MixServiceTest : AbstractExternalModelServiceTest<Mix, MixSaveDto, MixUpdateDto, MixService, MixRepository>() {
    override val repository: MixRepository = mock()
    private val recipeService: RecipeService = mock()
    private val materialTypeService: MaterialTypeService = mock()
    private val mixMaterialService: MixMaterialService = mock()
    private val mixTypeService: MixTypeService = mock()
    override val service: MixService =
      spy(MixServiceImpl(repository, recipeService, materialTypeService, mixMaterialService, mixTypeService))

    override val entity: Mix = mix(id = 0L, location = "location")
    override val anotherEntity: Mix = mix(id = 1L)
    override val entitySaveDto: MixSaveDto =
      spy(mixSaveDto(mixMaterials = setOf(mixMaterialDto(materialId = 1L, quantity = 1000f, position = 0))))
    override val entityUpdateDto: MixUpdateDto = spy(mixUpdateDto(id = entity.id!!))

    @AfterEach
    override fun afterEach() {
        super.afterEach()
        reset(recipeService, materialTypeService, mixMaterialService, mixTypeService)
    }

    // getAllByMixType()

    @Test
    fun `getAllByMixType() returns all mixes with the given mix type`() {
        val mixType = mixType(id = 0L)

        whenever(repository.findAllByMixType(mixType)).doReturn(entityList)

        val found = service.getAllByMixType(mixType)

        assertEquals(entityList, found)
    }

    // save()

    @Test
    override fun `save(dto) calls and returns save() with the created entity`() {
        val recipe = recipe(id = entitySaveDto.recipeId)
        val materialType = materialType(id = entitySaveDto.materialTypeId)
        val material = material(
          name = entitySaveDto.name,
          inventoryQuantity = Float.MIN_VALUE,
          isMixType = true,
          materialType = materialType
        )
        val mixType = mixType(name = entitySaveDto.name, material = material)
        val mix = mix(recipe = recipe, mixType = mixType)
        val mixWithId = mix(id = 0L, recipe = recipe, mixType = mixType)
        val mixMaterials = setOf(mixMaterial(material = material(id = 1L), quantity = 1000f))

        whenever(recipeService.getById(recipe.id!!)).doReturn(recipe)
        whenever(materialTypeService.getById(materialType.id!!)).doReturn(materialType)
        whenever(mixMaterialService.create(entitySaveDto.mixMaterials!!)).doReturn(mixMaterials)
        whenever(
          mixTypeService.getOrCreateForNameAndMaterialType(
            mixType.name,
            mixType.material.materialType!!
          )
        ).doReturn(mixType)
        doReturn(true).whenever(service).existsById(mixWithId.id!!)
        doReturn(mixWithId).whenever(service).save(any<Mix>())

        val found = service.save(entitySaveDto)

        verify(service).save(argThat<Mix> { this.recipe == mix.recipe })
        verify(recipeService).addMix(recipe, mixWithId)

        // Verify if this method is called instead of the MixType's constructor, which does not check if the name is already taken by a material.
        verify(mixTypeService).getOrCreateForNameAndMaterialType(mixType.name, mixType.material.materialType!!)

        assertEquals(mixWithId, found)
    }

    // update()

    private fun mixUpdateDtoTest(
      scope: MixUpdateDtoTestScope = MixUpdateDtoTestScope(),
      sharedMixType: Boolean = false,
      op: MixUpdateDtoTestScope.() -> Unit
    ) {
        with(scope) {
            doReturn(true).whenever(service).existsById(mix.id!!)
            doReturn(mix).whenever(service).getById(mix.id!!)
            doReturn(sharedMixType).whenever(service).mixTypeIsShared(mix.mixType)
            doAnswer { it.arguments[0] }.whenever(service).update(any<Mix>())

            if (mixUpdateDto.materialTypeId != null) {
                whenever(materialTypeService.getById(materialType.id!!)).doReturn(materialType)
            }

            op()
        }
    }

    private fun mixUpdateDtoMixTypeTest(sharedMixType: Boolean = false, op: MixUpdateDtoTestScope.() -> Unit) {
        with(MixUpdateDtoTestScope(mixUpdateDto = mixUpdateDto(id = 0L, name = "name", materialTypeId = 0L))) {
            mixUpdateDtoTest(this, sharedMixType, op)
        }
    }

    @Test
    override fun `update(dto) calls and returns update() with the created entity`() {
        val mixUpdateDto = spy(mixUpdateDto(id = 0L, name = null, materialTypeId = null))

        doReturn(entity).whenever(service).getById(any())
        doReturn(entity).whenever(service).update(entity)

        val found = service.update(mixUpdateDto)

        verify(service).update(entity)

        assertEquals(entity, found)
    }

    @Test
    fun `update(dto) calls MixTypeService saveForNameAndMaterialType() when mix type is shared`() {
        mixUpdateDtoMixTypeTest(sharedMixType = true) {
            whenever(mixTypeService.saveForNameAndMaterialType(mixUpdateDto.name!!, materialType))
              .doReturn(newMixType)

            val found = service.update(mixUpdateDto)

            verify(mixTypeService).saveForNameAndMaterialType(mixUpdateDto.name!!, materialType)

            assertEquals(newMixType, found.mixType)
        }
    }

    @Test
    fun `update(dto) calls MixTypeService updateForNameAndMaterialType() when mix type is not shared`() {
        mixUpdateDtoMixTypeTest {
            whenever(mixTypeService.updateForNameAndMaterialType(mixType, mixUpdateDto.name!!, materialType))
              .doReturn(newMixType)

            val found = service.update(mixUpdateDto)

            verify(mixTypeService).updateForNameAndMaterialType(mixType, mixUpdateDto.name!!, materialType)

            assertEquals(newMixType, found.mixType)
        }
    }

    @Test
    fun `update(dto) update, create and delete mix materials according to the given mix materials map`() {
        mixUpdateDtoTest {
            val mixMaterials = setOf(
              mixMaterialDto(materialId = 0L, quantity = 100f, position = 0),
              mixMaterialDto(materialId = 1L, quantity = 200f, position = 1),
              mixMaterialDto(materialId = 2L, quantity = 300f, position = 2),
              mixMaterialDto(materialId = 3L, quantity = 400f, position = 3),
            )
            mixUpdateDto.mixMaterials = mixMaterials

            whenever(mixMaterialService.create(any<Set<MixMaterialDto>>())).doAnswer {
                (it.arguments[0] as Set<MixMaterialDto>).map { dto ->
                    mixMaterial(
                      material = material(id = dto.materialId),
                      quantity = dto.quantity,
                      position = dto.position
                    )
                }.toSet()
            }

            val found = service.update(mixUpdateDto)

            mixMaterials.forEach {
                assertTrue {
                    found.mixMaterials.any { mixMaterial ->
                        mixMaterial.material.id == it.materialId && mixMaterial.quantity == it.quantity && mixMaterial.position == it.position
                    }
                }
            }
        }
    }

    // updateLocations()

    @Test
    fun `updateLocations() calls updateLocation() for each given MixLocationDto`() {
        val locations = setOf(
          mixLocationDto(mixId = 0, location = "Loc 0"),
          mixLocationDto(mixId = 1, location = "Loc 1"),
          mixLocationDto(mixId = 2, location = "Loc 2"),
          mixLocationDto(mixId = 3, location = "Loc 3")
        )

        service.updateLocations(locations)

        locations.forEach {
            verify(service).updateLocation(it)
        }
    }

    // updateLocation()

    @Test
    fun `updateLocation() updates the location of a mix in the repository according to the given MixLocationDto`() {
        val locationDto = mixLocationDto(mixId = 0L, location = "Location")

        service.updateLocation(locationDto)

        verify(repository).updateLocationById(locationDto.mixId, locationDto.location)
    }

    // delete()

    override fun `delete() deletes in the repository`() {
        whenCanBeDeleted {
            super.`delete() deletes in the repository`()
        }
    }

    // deleteById()

    @Test
    override fun `deleteById() deletes the entity with the given id in the repository`() {
        whenCanBeDeleted {
            super.`deleteById() deletes the entity with the given id in the repository`()
        }
    }

    private fun whenCanBeDeleted(id: Long = any(), test: () -> Unit) {
        whenever(repository.canBeDeleted(id)).doReturn(true)

        test()
    }
}

data class MixUpdateDtoTestScope(
  val mixType: MixType = mixType(name = "mix type"),
  val newMixType: MixType = mixType(name = "another mix type"),
  val materialType: MaterialType = materialType(id = 0L),
  val mix: Mix = mix(id = 0L, mixType = mixType),
  val mixUpdateDto: MixUpdateDto = spy(
    mixUpdateDto(
      id = 0L,
      name = null,
      materialTypeId = null,
      mixMaterials = setOf()
    )
  )
)
